/*
** Copyright (C) 2004-2024 by Carnegie Mellon University.
**
** @OPENSOURCE_LICENSE_START@
**
** SiLK 3.23
**
** Copyright 2024 Carnegie Mellon University.
**
** NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
** INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
** UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
** IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
** FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
** OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
** MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
** TRADEMARK, OR COPYRIGHT INFRINGEMENT.
**
** Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
** contact permission@sei.cmu.edu for full terms.
**
** [DISTRIBUTION STATEMENT A] This material has been approved for public
** release and unlimited distribution.  Please see Copyright notice for
** non-US Government use and distribution.
**
** This Software includes and/or makes use of Third-Party Software each
** subject to its own license.
**
** DM24-1064
**
** @OPENSOURCE_LICENSE_END@
*/

/*
**
**  Exports a method to portably set socket send/receive buffer sizes.
**
*/


#include <silk/silk.h>

RCSIDENT("$SiLK: sku-bigsockbuf.c 6c9dfbe0b623 2024-01-17 18:02:34Z mthomas $");

#include <silk/utils.h>

#ifndef GSB_DEBUG
#define GSB_DEBUG 0
#endif

/*
 * function: skGrowSocketBuffer
 *
 * There is no portable way to determine the max send and receive buffers
 * that can be set for a socket, so guess then decrement that guess by
 * 2K until the call succeeds.  If n > 1MB then the decrement by .5MB
 * instead.
 *
 * returns size or -1 for error
 */
int
skGrowSocketBuffer(
    int                 fd,
    int                 dir,
    int                 n)
{
#if GSB_DEBUG
    /* remember initial size */
    int size = n;
    int tries = 0;
#endif

    while (n > 4096) {
        if (setsockopt(fd, SOL_SOCKET, dir, (char*)&n, sizeof(n)) == 0) {
            return n;
        }
        /* anything other than no buffers available is fatal */
        if (errno != ENOBUFS) {
            return -1;
        }
        /* try a smaller value */
        if (n > 1024*1024) { /*most systems not > 256K bytes w/o tweaking*/
            n -= 1024*512;
        } else {
            n -= 2048;
        }
#if GSB_DEBUG
        ++tries;
#endif
    } /* while */

    /* no increase in buffer size */
    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
